package external;

import java.util.Arrays;

/**
 * @author Stephen Tavener
 *
 */
public class NoughtsAndCrosses implements ExternalGame 
{
	private static final String[] ROW_LABELS = { "1", "2", "3" }; 
	private static final String[] COL_LABELS = { "a", "b", "c" }; 

	private static final int OHS = 1;
	private static final int XES = 2;

	/** Board, as 2d array; rows, then columns */ 
	private int[][] board;
	
	/** Because noughts and crosses is a strict placement game, I know there will be exactly 9 moves in the game, each to an empty cell ... */
	private int moveCount;
	
	/** To mark if someone has won the game */
	private int winner;
	
	/**
	 * Default constructor - required!
	 */
	public NoughtsAndCrosses () 
	{
		// No action required
	}
	
	/**
	 * Copy constructor
	 * @param that
	 */
	public NoughtsAndCrosses (NoughtsAndCrosses that) 
	{
		// Copy the board cells - make sure we do a deep copy...
		this.board = new int[that.board.length][];
		for (int i = 0; i < board.length; i++)
			this.board[i] = that.board[i].clone();
		
		this.moveCount = that.moveCount;
		this.winner =  that.winner;
	}
	

	@Override
	public void init (String json) 
	{
		// One off initialisation
		board = new int[3][3];
		moveCount = 9;
		winner = 0;
	}

	@Override
	public int numMoves() {
		// moveCount is equal to the number of empty spaces and therefore the number of moves
		return moveCount;
	}

	@Override
	public void applyNthMove(int n) 
	{
		// Need to find the nth move, i.e. the nth empty space...
		int idx = 0;
		for (int r = 0; r < board.length; r++)
			for (int c = 0; c < board.length; c++)
				if (board[r][c] == 0)	// Empty?
					if (idx++ == n) 
					{
						// Found it!
						// Update the board
						board[r][c] = getOwner();
						
						// Reduce the number of moves/empty spaces reamining
						moveCount--;
						
						// Was this the winning move?
						checkWinner(board[r][c]);
						
						return;
					}
	}

	@Override
	public String actionDescription(int n) 
	{
		// Need to find the nth move, i.e. the nth empty space...
		int idx = 0;
		for (int r = 0; r < board.length; r++)
			for (int c = 0; c < board.length; c++)
				if (board[r][c] == 0)	// Empty?
					if (idx++ == n) {
						return COL_LABELS[c]+ROW_LABELS[r];
					}
		return "Error!";	// Should never get here
	}
	
	private void checkWinner(int lastPlayer) 
	{
		// There are only 8 lines through the board - check them all
		// In fact, we only need to check the ones that include the last stone played
		// If this were a more complex game, I'd take advantage of that to speed up processing
		// I'd also note that no-one can make a line until the 5th move and avoid unnecessary checking
		if (
				board[0][0] == lastPlayer && board[0][1] == lastPlayer && board[0][2] == lastPlayer ||
				board[1][0] == lastPlayer && board[1][1] == lastPlayer && board[1][2] == lastPlayer ||
				board[2][0] == lastPlayer && board[2][1] == lastPlayer && board[2][2] == lastPlayer ||

				board[0][0] == lastPlayer && board[1][0] == lastPlayer && board[2][0] == lastPlayer ||
				board[0][1] == lastPlayer && board[1][1] == lastPlayer && board[2][1] == lastPlayer ||
				board[0][2] == lastPlayer && board[1][2] == lastPlayer && board[2][2] == lastPlayer ||

				board[0][0] == lastPlayer && board[1][1] == lastPlayer && board[2][2] == lastPlayer ||
				board[0][2] == lastPlayer && board[1][1] == lastPlayer && board[2][0] == lastPlayer
			) 
		{
			winner = lastPlayer;
		}

		// No winner, no movews left => draw
		if (winner == 0 && moveCount == 0) winner = -1;
	}

	@Override
	public int getOwner() {
		// Player to move can be inferred from the number of empty spaces
		
		if ((moveCount&1)==0) return XES; // Even move? 'X'
		return OHS;	// Odd move => 'O'
	}

	@Override
	public boolean isTerminal() {
		// Has the game finished?  If so, winner will be set to that value or -1 if everyone draws
		return winner != 0;
	}

	@Override
	public int mcScore(int player) {
		// Calculate score for this player at this stage in the game
		// This function should only be called if isTerminal returns true
		if (winner < 0) return ExternalGame.DRAW;
		if (winner == player) return ExternalGame.WIN;
		return ExternalGame.LOSS;
	}

	@Override
	public ExternalGame deepCopy() {
		return new NoughtsAndCrosses(this);
	}

	@Override
	public long hash(int moveIndex) 
	{
		 /* Returns the hash for the current position if moveIndex is -1, index of specified move otherwise
		 * The game MUST return a hash for the current position when requested, but may return -1 for any other position; 		
		 */
		if (moveIndex >= 0) return -1;
		
		// Normally I'd use a Zobrist hash https://en.wikipedia.org/wiki/Zobrist_hashing
		// However, this whole game fits into a very small amount of bits (3^10 = Around 12 bits), so can use the exact board state as a hash
		long hash = getOwner();
		for (int r = 0; r < board.length; r++)
			for (int c = 0; c < board.length; c++)
			{
				hash *= 3;
				hash += board[r][c];
			}
		
		return hash;
	}


	@Override
	public boolean isHeuristic() {
		// Not using heuristics - we will rely on UCT for the search 
		return false;
	}

	@Override
	public int[] calcHeuristics() {
		// Not using heuristics, ignore this 
		return null;
	}

	@Override
	public void nullMove() {
		// Not using heuristics, ignore this 
	}

	@Override
	public boolean isStable() {
		// Not using heuristics, ignore this 
		return true;
	}

	@Override
	public int[] recommendedMoveOrder() {
		// External manager will return a random ordering if we return null 
		// That's usually what we want
		return null;
	}

	@Override
	public String[] drawables() {
		// Allocate enough space for all moves....
		final String[] results = new String[9];

		int idx = 0;
		for (int r = 0; r < board.length; r++)
			for (int c = 0; c < board.length; c++) 
			{
				switch (board[r][c]) 
				{
					case OHS: 
						results[idx++] = "stone" + DRAWABLE_SEPARATOR + "White" + DRAWABLE_SEPARATOR + COL_LABELS[c]+ROW_LABELS[r]; 
						break;
					case XES: 
						results[idx++] = "stone" + DRAWABLE_SEPARATOR + "Black" + DRAWABLE_SEPARATOR + COL_LABELS[c]+ROW_LABELS[r]; 
						break;
					default:
						// Nothing to display
						break;
				}
			}
		
		// reduce array to the correct size
		return Arrays.copyOf(results, idx);
	}

	@Override
	public String stateDescription() {
		if (winner < 0) return "Draw!";
		if (winner == OHS) return "White won";
		if (winner == XES) return "Black won";
		if (getOwner()==OHS) return "White to play";
		return "Black to play";
	}

}
